/*
 * This file is part of CutAndPlay.
 *
 * Copyright (C) 2009 Stefan Westerfeld
 * Copyright (C) 2009 Treuwert Computer GmbH
 *
 * CutAndPlay is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package cutandplay;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.image.BufferedImage;
import java.awt.image.BufferedImageOp;

import javax.swing.JLayeredPane;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * @author Christophe Le Besnerais 
 */
public class ImageCropper extends JLayeredPane {
    
    private BufferedImage image;
    private BufferedImage thumbnail;
    private JPanel background;
    private JPanel croppingPanel;
    private JPanel resizePanel;
    private Rectangle crop;
    private float zoom = 1.0f;
    private int wheelZoomAmount = 5;
    private int aspect_x;
    private int aspect_y;
    private BufferedImageOp filter;
    private transient ChangeEvent changeEvent;

    public ImageCropper(final BufferedImage image, int aspect_x, int aspect_y) {
	this.aspect_x = aspect_x;
	this.aspect_y = aspect_y;

        InnerListener listener = new InnerListener();
        this.addComponentListener(listener);
        this.image = image;
        this.crop = new Rectangle(image.getWidth() / 5, image.getHeight() / 5, 3 * image.getWidth() / 5, 3 * image.getHeight() / 5);
	constrainAspect (true);
        
        this.background = new JPanel(true) {            
            @Override
            protected void paintComponent(Graphics g) {
                g.drawImage(getThumbbail(), (this.getWidth() - getThumbbail().getWidth()) / 2, (this.getHeight() - getThumbbail().getHeight()) / 2, null);
            }        
        };
        this.add(background, JLayeredPane.DEFAULT_LAYER);
        
        this.croppingPanel = new JPanel(true) {
            @Override
            protected void paintComponent(Graphics g) {
                Graphics2D g2 = (Graphics2D) g;
                float dash1[] = { 10.0f };
                BasicStroke dashed = new BasicStroke(1.0f,
                         BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, dash1, 0.0f);
		int sx = crop.x * getThumbbail().getWidth() / image.getWidth();
		int sy = crop.y * getThumbbail().getHeight() / image.getHeight();
		int sx2 = sx + crop.width * getThumbbail().getWidth() / image.getWidth();
		int sy2 = sy + crop.height * getThumbbail().getHeight() / image.getHeight();
		g2.drawImage (getThumbbail(), 0, 0, this.getWidth(), this.getHeight(),
			                      sx, sy, sx2, sy2, null);
		//g2.fillRect (0, 0, this.getWidth(), this.getHeight());
                g2.setColor(Color.BLACK);
                g2.drawRect(0, 0, this.getWidth() - 1, this.getHeight() - 1);
                g2.setColor (Color.YELLOW);
                g2.setStroke (dashed);
                g2.drawRect(0, 0, this.getWidth() - 1, this.getHeight() - 1);
            }
        };
        this.croppingPanel.addMouseListener(listener);
        this.croppingPanel.addMouseMotionListener(listener);
        this.croppingPanel.addMouseWheelListener(listener);
        this.croppingPanel.setCursor(Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR));
        this.add(croppingPanel, JLayeredPane.PALETTE_LAYER, 2);
        
        this.resizePanel = new JPanel(true) {        
            @Override
            protected void paintComponent(Graphics g) {
                g.setColor(Color.WHITE);
                g.fillRect(0, 0, this.getWidth(), this.getHeight());
                g.setColor(Color.BLACK);
                g.drawRect(0, 0, this.getWidth() - 1, this.getHeight() - 1);
                g.drawLine(2, 4, 4, 2);
                g.drawLine(2, 7, 7, 2);
                g.drawLine(5, 7, 7, 5);
            }
        };
        this.resizePanel.addMouseListener(listener);
        this.resizePanel.addMouseMotionListener(listener);
        this.resizePanel.setCursor(Cursor.getPredefinedCursor(Cursor.NW_RESIZE_CURSOR));
        this.resizePanel.setSize(10, 10);
        this.add(resizePanel, JLayeredPane.PALETTE_LAYER, 0);
    }

    public void constrainAspect (boolean min)
    {
	int h1 = crop.height;
	int w1 = crop.height / aspect_y * aspect_x;
	int w2 = crop.width;
	int h2 = crop.width / aspect_x * aspect_y;
	int xmin = min?-1:1;
	if (h1 * w1 * xmin > h2 * w2 * xmin)
	{
	    crop.height = h1;
	    crop.width = w1;
	}
	else
	{
	    crop.height = h2;
	    crop.width = w2;
	}
    }
    public BufferedImageOp getFilter() {
        return filter;
    }

    public void setFilter(BufferedImageOp filter) {
        this.filter = filter;
    }
    
    public Rectangle getCropArea() {
	return crop;
    }

    private BufferedImage getThumbbail() {
        if (thumbnail == null) {
            int width = ImageCropper.this.image.getWidth();
            int height = ImageCropper.this.image.getHeight();
            float ratio = (float) width / (float) height;
            if ((float) this.getWidth() / (float) this.getHeight() > ratio) {
                height = this.getHeight();
                width = (int) (height * ratio);
            } else {
                width = this.getWidth();
                height = (int) (width / ratio);
            }
            
            thumbnail = getGraphicsConfiguration().createCompatibleImage(width, height, ImageCropper.this.image.getTransparency());
            Graphics2D g2d = thumbnail.createGraphics();
            //g2d.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR);
            g2d.drawImage(ImageCropper.this.image, 0, 0, width, height, null);
            g2d.dispose();
            
            if (filter != null)
                thumbnail = filter.filter(thumbnail, null);
        }
        
        return thumbnail;
    }
    
    private void changePosition() {
        if (crop.x < 0)
            crop.x = 0;
        if (crop.x > image.getWidth())
            crop.x = image.getWidth();
        if (crop.y < 0)
            crop.y = 0;
        if (crop.y > image.getHeight())
            crop.y = image.getHeight();
        if (crop.x + crop.width > image.getWidth()) 
            crop.x = image.getWidth() - crop.width;
        if (crop.y + crop.height > image.getHeight()) 
            crop.y = image.getHeight() - crop.height;
        
        changeCrop(true);
    }
    
    private void changeSize() {
        if (crop.width <= 0)
            crop.width = 1;
        if (crop.height <= 0)
            crop.height = 1;
        float ratio = (float) crop.width / (float) crop.height;
        if (crop.x + crop.width > image.getWidth()) {
            crop.width = image.getWidth() - crop.x;
            crop.height = (int) (crop.width / ratio);
        }
        if (crop.y + crop.height > image.getHeight()) {
            crop.height = image.getHeight() - crop.y;
            crop.width = (int) (crop.height * ratio);
        }
        if (crop.width >= image.getWidth()) {
            crop.width = image.getWidth();
            crop.height = (int) (crop.width / ratio);
        }
        if (crop.height >= image.getHeight()) {
            crop.height = image.getHeight();
            crop.width = (int) (crop.height * ratio);
        }
        
        changeCrop(true);
    }
    
    private void changeZoom() {
        if (crop.x < 0)
            crop.x = 0;
        if (crop.y < 0)
            crop.y = 0;
        if (crop.width <= 0)
            crop.width = 1;
        if (crop.height <= 0)
            crop.height = 1;
        float ratio = (float) crop.width / (float) crop.height;
        if (crop.x + crop.width > image.getWidth()) {
            crop.width = image.getWidth() - crop.x;
            crop.height = (int) (crop.width / ratio);
        }
        if (crop.y + crop.height > image.getHeight()) {
            crop.height = image.getHeight() - crop.y;
            crop.width = (int) (crop.height * ratio);
        }
        if (crop.width >= image.getWidth()) {
            crop.width = image.getWidth();
            crop.height = (int) (crop.width / ratio);
        }
        if (crop.height >= image.getHeight()) {
            crop.height = image.getHeight();
            crop.width = (int) (crop.height * ratio);
        }
        
        changeCrop(false);
    }
    
    private void changeCrop(boolean updatePanel) {
        this.fireStateChanged();

        if (updatePanel) {
            float ratio = (float) image.getWidth() / (float) getThumbbail().getWidth();
            this.croppingPanel.setSize((int) (zoom * crop.width / ratio), (int) (zoom * crop.height / ratio));
            this.croppingPanel.setLocation((int) ((crop.x / ratio) - ((crop.width / ratio) * (zoom - 1) / 2)) + (getWidth() - getThumbbail().getWidth()) / 2, (int) ((crop.y / ratio) - ((crop.height / ratio) * (zoom - 1) / 2)) + (getHeight() - getThumbbail().getHeight()) / 2);
            this.resizePanel.setLocation(croppingPanel.getX() + croppingPanel.getWidth() - resizePanel.getWidth() / 2, croppingPanel.getY() + croppingPanel.getHeight() - resizePanel.getHeight() / 2);
        } else {
            this.croppingPanel.repaint();
        }
    }
    
    public void addChangeListener(ChangeListener l) {
        listenerList.add(ChangeListener.class, l);
    }
    
    public void removeChangeListener(ChangeListener l) {
        listenerList.remove(ChangeListener.class, l);
    }
    
    public ChangeListener[] getChangeListeners() {
        return (ChangeListener[])(listenerList.getListeners(
            ChangeListener.class));
    }

    protected void fireStateChanged() {
        Object[] listeners = listenerList.getListenerList();
        for (int i = listeners.length-2; i>=0; i-=2) {
            if (listeners[i]==ChangeListener.class) {
                if (changeEvent == null)
                    changeEvent = new ChangeEvent(this);
                ((ChangeListener)listeners[i+1]).stateChanged(changeEvent);
            }          
        }
    }   
    
    
    private class InnerListener extends MouseAdapter implements ComponentListener, MouseMotionListener, MouseWheelListener {

        private Point cursorPosition;
        
        public void componentResized(ComponentEvent e) {
            thumbnail = null;
            background.setSize(ImageCropper.this.getSize());
            float ratio = (float) getThumbbail().getWidth() / (float) image.getWidth();
            croppingPanel.setSize((int) (zoom * crop.width * ratio), (int) (zoom * crop.height * ratio));
            croppingPanel.setLocation((int) ((crop.x * ratio) - (crop.width * ratio * (zoom - 1) / 2)) + (getWidth() - getThumbbail().getWidth()) / 2, (int) ((crop.y * ratio) - (crop.height * ratio * (zoom - 1) / 2)) + (getHeight() - getThumbbail().getHeight()) / 2);
            resizePanel.setLocation(croppingPanel.getX() + croppingPanel.getWidth() - resizePanel.getWidth() / 2, croppingPanel.getY() + croppingPanel.getHeight() - resizePanel.getHeight() / 2);
            thumbnail = null;
        }
        
        @Override
        public void mousePressed(MouseEvent e) {
            cursorPosition = e.getPoint();
            setLayer(e.getComponent(), JLayeredPane.DRAG_LAYER);
        }

        @Override
        public void mouseReleased(MouseEvent e) {
            cursorPosition = null;
            setLayer(e.getComponent(), JLayeredPane.PALETTE_LAYER);
            setPosition(croppingPanel, 1);
            setPosition(resizePanel, 0);
        }

        public void mouseDragged(MouseEvent e) {
            if (cursorPosition != null) {
                if (e.getComponent().equals(croppingPanel)) {
                    float ratio = (float) image.getWidth() / (float) getThumbbail().getWidth();
                    Point pos = SwingUtilities.convertPoint(croppingPanel, e.getPoint(), background);
                    pos.translate(- cursorPosition.x, - cursorPosition.y);
                    pos.translate(croppingPanel.getWidth() / 2, croppingPanel.getHeight() / 2);
                    pos.translate(- (getWidth() - getThumbbail().getWidth()) / 2, - (getHeight() - getThumbbail().getHeight()) / 2); 
                    pos.setLocation(pos.x * ratio, pos.y * ratio);
                    pos.translate(- crop.width / 2, - crop.height / 2);
                    crop.setLocation(pos.x, pos.y);
                    changePosition();
                    
                } else if (e.getComponent().equals(resizePanel)) {
                    float ratio = (float) image.getWidth() / (float) getThumbbail().getWidth();
                    Point start = SwingUtilities.convertPoint(resizePanel, cursorPosition, background);
                    Point end = SwingUtilities.convertPoint(resizePanel, e.getPoint(), background);
                    crop.setSize(crop.width + (int) ((end.x - start.x) * ratio), crop.height + (int) ((end.y - start.y) * ratio));
		    constrainAspect (false);
                    changeSize();
                }
            }
        }
        
        public void mouseWheelMoved(MouseWheelEvent e) {
            if ((e.getWheelRotation() > 0 && crop.x > 0 && crop.x + crop.width < image.getWidth() && crop.y > 0 && crop.y + crop.height < image.getHeight())
                    || (e.getWheelRotation() < 0 && crop.width > 1 && crop.height > 1)) {
                float ratio = (float) crop.height / (float)crop.width;                
                int newWidth = (int) (crop.width + (e.getWheelRotation() * wheelZoomAmount));
                int newHeight = (int) (newWidth * ratio);
                crop.setBounds(crop.x - ((newWidth - crop.width) / 2), crop.y - ((newHeight - crop.height) / 2), newWidth, newHeight);
                zoom = (float) (croppingPanel.getWidth() * image.getWidth()) / (float) (crop.width * getThumbbail().getWidth());
                changeZoom();
            }
        }

        public void componentMoved(ComponentEvent e) {}
        public void componentShown(ComponentEvent e) {}
        public void componentHidden(ComponentEvent e) {}
        public void mouseMoved(MouseEvent e) {}        
        
    }

}

